(* Basic gates and circuits *)

(* Higher-kinded programming in OCaml, as described in
   Lightweight Higher-Kinded Polymorphism
   Jeremy Yallop and Leo White
   Functional and Logic Programming 2014
*)

(* 
#load "higher.cmo";;
*)
open Higher;;


(* Define the syntax of circuits (in the form close to BNF)
   we have NOT, AND, OR gates
 * Read  ``bool repr'' as a wire
   We have constant wires (grounded or connected to the power supply),
   we can invert a wire, we can connect two wires to the AND
   gate and get an output wire in return.
*)

module type SYM = sig
  type 'a repr
  type obs

  val lit : bool -> bool repr
  val neg : bool repr -> bool repr
  val an_ : bool repr -> bool repr -> bool repr
  val or_ : bool repr -> bool repr -> bool repr

  val observe : 'a repr -> ('a,obs) app
end

(* 
 The function observe with some unusual type will be explained later on.
 Just take it for granted for now. And also the following line: a part
 of a boilerplate.
*)
type 'obs sym = (module SYM with type obs = 'obs)

(*
   We can already build simple circuits
   I will stand for `interpreter', an instance of SYM
 *)

(* See inferred type, cf Haskell and complain about syntax and boilerplate *)
let ex1 = fun (type obs) ((module I):obs sym) -> let open I in observe @@
  an_ (lit true) (neg (lit false))

(* An alternative
   `Object algebras' (although it was presented at SSGIP in 2010,
   two years before object algebras)

class virtual ['repr,'obs] sym = object 
  method virtual lit : bool -> (bool,'repr) app
  method virtual neg : (bool,'repr) app -> (bool,'repr) app
  method virtual an_ : (bool,'repr) app -> 
                       (bool,'repr) app -> (bool,'repr) app
  method virtual or_ : ((bool,'repr) app as 'x) -> 'x -> 'x

  method virtual observe : 'a. ('a,'repr) app -> ('a,'obs) app
end

*)


(* Boring: all constants. 
   Need some inputs. So we extend the language
*)

module type SYMW = sig
  include SYM
  val wire_x : bool repr
  val wire_y : bool repr
  val wire_z : bool repr
end

type 'obs symw = (module SYMW with type obs = 'obs)

(*
class virtual ['repr,'obs] symw = object 
  inherit ['repr,'obs] sym
  method virtual wire_x : (bool,'repr) app
  method virtual wire_y : (bool,'repr) app
  method virtual wire_z : (bool,'repr) app
end
*)

(* check the inferred type! *)

let ex2 = fun (type obs) ((module I):obs symw) -> let open I in observe @@
  an_ (lit true) (neg wire_x)

(* How to get xor? Extend the language again?
   This is like a _macro_ (we see later)
*)
module XOR(I:SYM) = struct
  open I
  let xor x y = an_ (or_ x y) (neg (an_ x y))
end

(*
 * simple adder
*)

let exadd2 = fun (type obs) ((module I):obs symw) -> 
   let module X = XOR(I) in let open X in let open I in observe @@
     xor wire_x wire_y

let exadd3 = fun (type obs) ((module I):obs symw) -> 
   let module X = XOR(I) in let open X in let open I in observe @@
     xor (xor wire_x wire_y) wire_z

let exadd_xy1 = fun (type obs) ((module I):obs symw) -> 
   let module X = XOR(I) in let open X in let open I in observe @@
    xor wire_x (xor wire_y (lit true))

(* How can we check this is correct?
   We need an interpreter
 *)

(* The structure also defines denotational semantics! Compositional
   (as denotational semantics should be)
   The meaning of an expression is defined entirely through
   the meaning (but not form) of the immediate components
   That is why it is called Symantics
*)


(* // *)
(* Meta-circular interpreter (usually called R)
   It interprets |'a repr| as just |a|, that is, |bool repr|
   as just OCaml boolean
    R is a record with one field, called eval
     data R a = R{eval:: a}
*)

(* consider an iso-morphism between |'a r| and |('a,rname) app|
 * where
 *  type 'a r = 'a
 *
 *  val inj_r : 'a r -> ('a,rname) app
 *  val prj_r : ('a,rname) app -> 'a r
 *
 * Why we need the iso-morphism? Emulation of higher-kinded polymorphism
 * (see the beginning of this file). 
 *  So, rname is the name of the type constructor r.

 In Yallop and White's `higher', we do 
*)

module N_id = Newtype1(struct type 'a t = 'a end)
(* 
  so that inj_r is spelled N_id.inj and prj_r is spelled N_id.prj
  and rname is spelled N_id.t
*)

(* It is really the metacircular interpreter *)
module R = struct
  type 'a repr = 'a
  type obs = N_id.t
  let lit x = x
  let neg = not
  let an_ = (&&)
  let or_ = (||)

  let observe = N_id.inj 
end

  
let ex1_eval = N_id.prj @@ ex1 @@ (module R)


(* To run ex2, we need an instance of SYMW *)

module RW = struct
  include R
  let wire_x = true
  let wire_y = false
  let wire_z = true
end

(* What if we use a wrong interpreter

let exadd2_eval = N_id.prj @@ exadd2 @@ (module R)
*)

let exadd2_eval = N_id.prj @@ exadd2 @@ (module RW)



(* QUIZ *)
(* The SymW instance had constant values. What if we want different
   values?
 Compare with Haskell and discuss trade-offs
 *  -- keep writing interpreters?
 *  -- local instances? implicit parameters? dynamic binding!
 *  -- write a new interpreter
 *       data RV a = RV ((Bool,Bool,Bool) -> a)
 *     But we need to re-write Symantics
 *  -- How to write Symantics R so we don't have to change it
 *     when changing from R to RV
 * (Applicative)
*)



let exadd2_eval = N_id.prj @@ exadd2 @@ (module struct include RW
  let wire_x = false end)

 (* abstracting wire_x value *)
let exadd2_eval' x = N_id.prj @@ exadd2 @@ (module struct include RW
  let wire_x = x end)


(* // *)
(* Another interpreter, to show the circuit *)

(* 
 Interpret |a repr| as a string, regardless of the type |a|
*)

module N_string = Newtype1(struct type 'a t = string end)

module S0 = struct
  type 'a repr = string
  type obs = N_string.t

  let paren : string -> string = fun x -> "(" ^ x ^ ")"
 
  let lit x = if x then "tt" else "ff"
  let neg x   = "~" ^ x
  let an_ x y = paren @@ x ^ " && " ^ y
  let or_ x y = paren @@ x ^ " || " ^ y

  let observe = N_string.inj
end

let ex1S = N_string.prj @@ ex1 @@ (module S0)

module SW0 = struct
  include S0
  let wire_x = "X"
  let wire_y = "Y"
  let wire_z = "Z"
end

let exadd2_view = N_string.prj @@ exadd2 @@ (module SW0)
let exadd3_view = N_string.prj @@ exadd3 @@ (module SW0)

(* Is it clear that xor was a macro? *)

(* QUIZ: How to avoid too many parenthesis?
   Write a precedence printer
*)


module S = struct
  type prec = int
  type 'a repr = prec -> string
  type obs = N_string.t

  let paren = function
    | true -> fun x -> "(" ^ x ^ ")"
    | _    -> fun x -> x
 
  let lit x   = fun p -> if x then "tt" else "ff"
  let neg x   = fun p -> paren (p>10) @@ "~" ^ (x 11)
  let an_ x y = fun p -> paren (p>4) @@ x 4 ^ " && " ^ y 4
  let or_ x y = fun p -> paren (p>3) @@ x 3 ^ " || " ^ y 3

  let observe x = N_string.inj (x 0)
end

module SW = struct
  include S
  let wire_x = fun _ -> "X"
  let wire_y = fun _ -> "Y"
  let wire_z = fun _ -> "Z"
end

let view e = N_string.prj @@ e @@ ((module SW) : SW.obs symw)

let exadd3_view = view exadd3

(* Simplification is in order! *)

(* QUIZ *)
(* write an interpreter to draw the circuit as a diagram *)

(* QUIZ *)
(* write an interpreter to compute the latency (circuit switching delay) *)

(* How to implement the circuit if we only have NAND gates? 
 * (show two ways, one using normalization)
*)

module type NAND = sig
  type 'a repr
  type obs

  val lit  : bool -> bool repr
  val nand : bool repr -> bool repr -> bool repr

  val observe : 'a repr -> ('a,obs) app
end

type 'obs nand = (module NAND with type obs = 'obs)



module TNAND(I:NAND) = struct
  type 'a repr = 'a I.repr
  type obs = I.obs

  let lit = I.lit
  let neg x   = I.nand x x
  let an_ x y = neg (I.nand x y)
  (* neg (neg (or x y)) = neg (and (neg x) (neg y)) *)
  let or_ x y = I.nand (neg x) (neg y)

  let observe = I.observe
end

let ex1S = N_string.prj @@ ex1 @@ (module S)

let ex1_nand = fun (type obs) ((module I):obs nand) -> 
  ex1 (module TNAND(I))


module SN = struct
  include S
  let nand x y = fun p -> paren (p>3) @@ x 3 ^ " // " ^ y 3
end

let ex1S = N_string.prj @@ ex1_nand @@ (module SN)

(* What about the wires? 
 * Monkey-patching
*)

module type NANDW = sig
  include NAND
  val wire_x : bool repr
  val wire_y : bool repr
  val wire_z : bool repr
end

type 'obs nandw = (module NANDW with type obs = 'obs)

let exadd3_nand = fun (type obs) ((module I):obs nandw) -> 
  exadd3(module (struct include TNAND(I) 
                    let wire_x = I.wire_x 
                    let wire_y = I.wire_y 
                    let wire_z = I.wire_z end))

(*
 * more monkey-patching: cf. local instances
 *)

let exadd3_nand_view = N_string.prj @@ exadd3_nand @@ 
   (module (struct include SW let nand = SN.nand end))


;;
